import logging
import random
from typing import Dict, List, Set, Tuple

from gevent.lock import BoundedSemaphore
from locust import events, task

from performance_tests.common import (
    ProvisioningClient,
    ResponseError,
    TestData,
    UsedSchools,
    get_classes,
    get_provisioning_api_admins,
)

school_authorities: List[str] = []  # ["Traeger1", "Traeger"]
user_credentials: Dict[str, Tuple[str, str]] = {}  # [(username, password), ..]
id_broker_school_classes: Dict[str, Dict[str, Dict[str, str]]] = {}  # {sch_auth: {ou: {cl: uuid, ...}}}
NUM_SCHOOLS_PER_SCHOOL_AUTHORITY = 5  # loading all OUs from test DB takes to long
used_schools = UsedSchools()
deleted_classes: Dict[str, Dict[str, Set[str]]] = {}  # {sch_auth: {ou: {cl1, cl2, ..}}}
sem_deleted_classes = BoundedSemaphore()


@events.init.add_listener
def on_init(environment, **kwargs):
    data = TestData()
    school_authorities.extend(data.school_authorities)
    deleted_classes.update({s_a: {} for s_a in school_authorities})
    user_credentials.update(get_provisioning_api_admins(data))
    id_broker_school_classes.update(get_classes(data, NUM_SCHOOLS_PER_SCHOOL_AUTHORITY))


class ProvisioningAPIClient(ProvisioningClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.school_authority = random.choice(school_authorities)
        logging.info("Testing with school authority: %r", self.school_authority)
        self.auth_username, self.auth_password = user_credentials[self.school_authority]
        self.schools = list(id_broker_school_classes[self.school_authority].keys())
        self.school = self.get_fresh_school()

    def get_fresh_school(self) -> str:
        while True:
            school = random.choice(self.schools)
            if (self.school_authority, school) not in used_schools.schools:
                break
        used_schools.add(self.school_authority, school)
        deleted_classes[self.school_authority][school] = set()
        logging.info(
            "Testing with school %r incl. %d classes.",
            school,
            len(id_broker_school_classes[self.school_authority][school]),
        )
        return school

    @task
    def school_class_delete(self) -> None:
        while True:
            classes = id_broker_school_classes[self.school_authority][self.school]
            with sem_deleted_classes:
                if len(deleted_classes[self.school_authority][self.school]) == len(classes):
                    # we have deleted all classes from this school
                    self.school = self.get_fresh_school()
                    continue
            name, entry_uuid = random.choice(list(classes.items()))
            with sem_deleted_classes:
                if name in deleted_classes[self.school_authority][self.school]:
                    continue
                deleted_classes[self.school_authority][self.school].add(name)
            break
        try:
            super().school_class_remove(school_authority=self.school_authority, obj_id=entry_uuid)
        except ResponseError as exc:
            logging.error(
                "%s: school_authority=%r name=%r entry_uuid=%r -> %r",
                self.greenlet.name,
                self.school_authority,
                name,
                entry_uuid,
                exc,
            )


@events.test_stop.add_listener
def _(environment, **_kwargs):
    logging.info("Used schools: %r", sorted(f"{s_a}-{ou}" for s_a, ou in used_schools.schools))
    used_schools.persist()
